
import simpleOAuthPkg from "simple-oauth2";
const { AuthorizationCode } = simpleOAuthPkg;

import logger from "../logger.js";

import {
  formatPhoneNumber, isEmptyString, getDateFormated,
  filterPhoneNumberInput, isNumberString, logObject
} from "../ison24/appUtils.js";

import exapi_client_helper from "../exapi-client/helper.js";

import exapi_client_model from "../exapi-client/model.js";

/**
* OAuth2 configuration
*/
const CLIENT_ID = process.env.CLIENT_ID;
const CLIENT_SECRET = process.env.CLIENT_SECRET;

const appController = {};

appController.home = async (req, res) => {
  logger.info('/home ');

  const users = await exapi_client_model.getOAuthUsers('active');
  if (!users || users.length === 0) {
    res.render('home.hbs', {
      title: 'Home | IsOn24 ExAPI Client App',
      connection_info: 'No connection info available',
      connect: '',
      disconnect: 'hidden',
    });
    return;
  }

  const latestUser = users[0];
  logger.debug(JSON.stringify(latestUser));
  res.render('dashboard.hbs', {
    title: 'Home | IsOn24 ExAPI Client App',
    connection_info: `Connected to ${latestUser.business_name} (${latestUser.business_email})`,
    connect: 'hidden', // exapi_client_model.isConnected(),
    disconnect: '', // exapi_client_model.isDisconnected(),
    user: latestUser,
  })
};

appController.dashboard = appController.home;

appController.disconnect = async (req, res) => {
  logger.debug(req.originalUrl);

  const { clid, bid } = req.query || {};
  logger.debug(`clid: ${clid}, bid: ${bid}`);

  if (isEmptyString(clid) || isEmptyString(bid)) {
    res.redirect('/');
    return;
  }
  // Update self state to inactive
  await exapi_client_model.updateUserState(clid, bid, 'inactive');

  // Send a disconnect request to ExAPI
  const result = await exapi_client_helper.disconnect(clid, bid);

  res.redirect('/');
};

appController.searchBusiness = async (req, res) => {
  logger.debug(req.originalUrl);

  const { clid, bid, search_email } = req.body;
  logger.debug(`clid: ${clid}, bid: ${bid}`);
  logger.debug(`email: ${search_email}`);

  if (isEmptyString(search_email)) {
    res.redirect('/');
    return;
  }

  const results = await exapi_client_helper.searchBusinesses(clid, bid, { email: search_email });
  logger.debug(`results: ${JSON.stringify(results)}`);
  // const sample = [{"id":"3ad30e05-08ba-11ee-94ac-067be461895f","name":"Samvadi LLC","sPhones":["4087964894"],"emails":["jking@ison24.com"],"timezone":"America/Los_Angeles"},
  // {"id":"1d78fd5e-5071-11ee-94ac-067be461895f","name":"Green canyon appliance repair","sPhones":["npanxx1694"],"emails":["jking@ison24.com"],"timezone":"America/Los_Angeles"},
  // {"id":"49e6ec2b-6ec1-11ee-94ac-067be461895f","name":"John KIng @ Red canyon appliance repair","sPhones":["7812623692"],"emails":["jking@ison24.com"],"timezone":"America/Los_Angeles"}];
  const formattedResults = results.map((result) => {
    return {
      ...result,
      // sPhones: result.sPhones.map( (phone) => formatPhoneNumber(phone)),
      sPhone1: formatPhoneNumber(result.sPhones[0]),
    };
  });
  res.render('search-results.hbs', {
    title: 'Search Results | IsOn24 ExAPI Client App',
    search_query: { email: search_email },
    results: formattedResults,
    user: {
      client_id: clid,
      business_id: bid,
    }
  });
};

appController.send_message = async (req, res) => {
  logger.debug(req.originalUrl);

  try {
    const { clid, bid, contact_phone, message } = req.body;
    logger.debug(`clid: ${clid}, bid: ${bid}, contact_phone: ${contact_phone}, message: ${message}`);

    if (isEmptyString(contact_phone) || isEmptyString(message)) {
      res.status(200).send('Empty request');
      return;
    }

    const result = await exapi_client_helper.sendMessage(clid, bid,
      {
        ContactNumber: contact_phone,
        Content: message
      });

    logger.debug(`result: ${JSON.stringify(result)}`);

    res.status(200).send('Message sent');
  } catch (error) {
    logger.error(`sendMessage error: ${error.toString()}`);
    res.status(500).send('Error sending message');
  }
};

appController.conversations_get = async (req, res) => {
  logger.debug(req.originalUrl);

  const { clid, bid, message_count, use_get_api, client_tz, contact_id } = req.query || {};
  logger.debug(`clid: ${clid}, bid: ${bid}, client_tz: ${client_tz}, contact_id: ${contact_id}`);

  await appController.conversations(req, res, clid, bid, message_count, use_get_api, client_tz, contact_id);
};

appController.conversations_post = async (req, res) => {
  logger.debug(req.originalUrl);

  const { clid, bid, message_count, use_get_api, client_tz, contact_id } = req.body || {};
  logger.debug(`clid: ${clid}, bid: ${bid}, client_tz: ${client_tz}, contact_id: ${contact_id}`);

  await appController.conversations(req, res, clid, bid, message_count, use_get_api, client_tz, contact_id);
};

appController.conversations = async (req, res, clid, bid, message_count, use_get_api, client_tz, contact_id) => {
  logger.debug(req.originalUrl);
  logger.debug(`clid: ${clid}, bid: ${bid}, client_tz: ${client_tz}, contact_id: ${contact_id}`);

  if(! isNumberString(message_count)) {
    message_count = 10;
  }

  let messages = [];
  if (use_get_api) {
    /*
    pull messages from ExAPI
    populate the messages page
    */
    messages = await exapi_client_helper.getMessages(clid, bid);
    logger.debug(`messages: ${JSON.stringify(messages)}`);
    res.render('messages.hbs', {
      title: 'Messages | IsOn24 ExAPI Client App',
      messages,
      user: {
        client_id: clid,
        business_id: bid,
      }
    });
    return;
  }
  /*
  pull messages from DB
  populate the messages page
  */
  let contacts = await exapi_client_model.getContacts(clid, bid, 10);
  logger.debug(`contacts: ${JSON.stringify(contacts)}`);
  messages = [];
  let selected_contact = {};

  if (contacts && contacts.length > 0) {
    // By default, select the first contact - Most recently updated
    selected_contact = contacts[0];
    if (contact_id) {
      selected_contact = contacts.find((contact) => contact._id === contact_id);
    }
    messages = await exapi_client_model.getMessagesForContact(clid, bid, selected_contact._id, message_count);
    logger.debug(`messages: ${JSON.stringify(messages)}`);
  }

  messages.forEach((message) => {
    message.createdAtFormated = getDateFormated(message.created_at, client_tz);
  });

  res.render('conversations.hbs', {
    title: 'Conversations | IsOn24 ExAPI Client App',
    contacts: contacts,
    selectedContact: selected_contact,
    messages,
    user: {
      client_id: clid,
      business_id: bid,
    }
  });
};

appController.appointments_get = async (req, res) => {
  logger.debug(req.originalUrl);

  const { clid, bid, appointment_count, use_get_api, client_tz } = req.query || {};
  logger.debug(`clid: ${clid}, bid: ${bid}`);

  await appController.appointments(req, res, clid, bid, appointment_count, use_get_api, client_tz);
};

appController.appointments_post = async (req, res) => {
  logger.debug(req.originalUrl);

  const { clid, bid, appointment_count, use_get_api, client_tz } = req.body;
  logger.debug(`clid: ${clid}, bid: ${bid}`);

  await appController.appointments(req, res, clid, bid, appointment_count, use_get_api, client_tz);
};

appController.appointments = async (req, res, clid, bid, appointment_count, use_get_api, client_tz) => {
  logger.debug(req.originalUrl);

  logger.debug(`clid: ${clid}, bid: ${bid}`);
  let count = 10;
  if (appointment_count) {
    count = appointment_count;
  }
  let appointments = [];
  if (use_get_api) {
    /*
    pull appointments from ExAPI
    Save appointments to DB
    */
    appointments = await exapi_client_helper.getAppointments(clid, bid, count);
    logger.debug(`appointments: ${JSON.stringify(appointments)}`);

    const result = await appointments.reduce(async (carry, appointment) => {
      return [
        ...(await carry),
        // Save received appointment update to DB
        await exapi_client_model.updateAppointment(
          clid, bid,
          appointment),
      ]
    }, Promise.resolve([]));
  }

  renderCalendarPage(req, res, clid, bid, count, client_tz);
};

appController.cancel_appointment = async (req, res) => {
  logger.debug(req.originalUrl);
  try {
    let { clid, bid, appt_id, client_tz } = req.body;
    logger.debug(`clid: ${clid}, bid: ${bid}`);
    logger.debug(`appt_id: ${appt_id}`);

    if (isEmptyString(appt_id)) {
      res.status(200).send('Empty request');
      return;
    }

    // Remove _cancel from appt_id
    appt_id = appt_id.replace('_cancel', '');

    let appointment = {};
    /*
    Save updates to appointments to DB
    */
    appointment = await exapi_client_model.cancelAppointment(clid, bid, { _id: appt_id, State: 'CANCELLED' });
    if (!appointment) {
      throw new Error(`Appointment ${appt_id} not found`);
    }
    logger.debug(`appointment: ${JSON.stringify(appointment)}`);

    // Fill contact info
    let contact = await exapi_client_model.getContact(clid, bid, { _id: appointment.contact_id });

    // Send updated appointment to IsOn24
    await exapi_client_helper.cancelAppointment(clid, bid,
      {
        ...appointment,
        phone: contact.phone,
      });

    // Refresh appointments page
    renderCalendarPage(req, res, clid, bid, 10, client_tz);
  } catch (error) {
    logger.error(`cancel_appointment error: ${error.toString()}`);
    res.status(500).send('Error cancelling appointment');
  }
};

const renderCalendarPage = async (req, res, clid, bid, max_count, client_tz) => {
  /*
  pull appointments from DB
  populate the appointments page
  */
  let contacts = await exapi_client_model.getContacts(clid, bid, 10);
  logger.debug(`contacts: ${JSON.stringify(contacts)}`);
  let appointments = [];
  // const tz_str = 'America/Los_Angeles';

  if (contacts && contacts.length > 0) {
    let contacts_map = new Map();
    contacts.forEach((contact) => {
      contacts_map.set(contact._id, contact);
    });
    appointments = await exapi_client_model.getAppointments(clid, bid, max_count);
    let f_appointments = appointments.map((appointment) => {
      if (!appointment.start) {
        return null;
      }
      // logger.debug(`typeof appointment.start: ${typeof appointment.start}`);
      return {
        ...appointment,
        startAt: getDateFormated(new Date(appointment.start), client_tz),
        duration: (Date.parse(appointment.end) - Date.parse(appointment.start)) / 1000 / 60,
        contactNumber: formatPhoneNumber(contacts_map.get(appointment.contact_id).phone),
        contactName: contacts_map.get(appointment.contact_id).first_name,
        contactEmail: contacts_map.get(appointment.contact_id).email,
      };
    });
    appointments = f_appointments;
    logger.debug(`appointments: ${JSON.stringify(appointments)}`);
  }

  res.render('calendar.hbs', {
    title: 'Calendar | IsOn24 ExAPI Client App',
    contacts: contacts,
    appointments,
    // timezone: tz_str,
    user: {
      client_id: clid,
      business_id: bid,
    }
  });
};

appController.create_appointment_post = async (req, res) => {
  logger.debug(req.originalUrl);

  let return_obj = { title: 'New Appointment | IsOn24 ExAPI Client App', };

  try {
    const { clid, bid,
      contact_phone, contact_name, contact_email,
      date, time, duration, start_ts,
      type, guest_count, content,
      use_get_api, client_tz } = req.body;
    logger.debug(`clid: ${clid}, bid: ${bid}`);
    return_obj.user = {
      client_id: clid,
      business_id: bid,
    };

    logger.debug(`contact_phone: ${contact_phone}, date: ${date}, time: ${time}, duration: ${duration}, start_ts: ${start_ts}`);

    if (isEmptyString(contact_phone) || isEmptyString(date) || isEmptyString(time) || isEmptyString(duration)) {
      throw new Error('Empty request');
    }

    let appointment = {};
    /*
    Save new appointment to DB
    call ExAPI to create appointment in IsOn24
    */
    // Save received appointment update to DB
    appointment = await exapi_client_model.createAppointment(
      clid, bid,
      {
        ContactNumber: filterPhoneNumberInput(contact_phone),
        ContactName: (isEmptyString(contact_name) ? null : contact_name),
        ContactEmail: (isEmptyString(contact_email) ? null : contact_email),

        StartTime: parseInt(start_ts),
        Duration: parseInt(duration),
        EndTime: parseInt(start_ts) + parseInt(duration) * 60 * 1000,

        State: 'CONFIRMED',
        EventType: type,
        GuestCount: (isNumberString(guest_count) ? parseInt(guest_count) : 1),
        Content: (content ? content : 'Appointment created by ExAPI Client App'),
        CustomInfo: '{}',
      });

    const appointment_update = {
      ...appointment,
      phone: filterPhoneNumberInput(contact_phone),
      name: (isEmptyString(contact_name) ? null : contact_name),
      email: (isEmptyString(contact_email) ? null : contact_email),
    };
    logger.debug(`appointment_update: ${JSON.stringify(appointment_update)}`);
    // Create appointment in IsOn24
    await exapi_client_helper.createAppointment(clid, bid,
      appointment_update);

    // Refresh appointments page
    // renderCalendarPage(req, res, clid, bid, 10);
    return_obj.result = 'Appointment created.';
  } catch (error) {
    return_obj.error = error.toString();
    logger.error(`create_appointment_post error: ${return_obj.error}`);
    // res.status(500).send('Error creating appointment');
  }
  logger.debug(`return_obj: ${JSON.stringify(return_obj)}`);
  res.render('new_appointment.hbs', return_obj);
};

appController.create_appointment_get = async (req, res) => {
  logger.debug(req.originalUrl);

  const { clid, bid, client_tz } = req.query || {};
  logger.debug(`clid: ${clid}, bid: ${bid}, client_tz: ${client_tz}`);

  if (isEmptyString(clid) || isEmptyString(bid)) {
    res.redirect('/');
    return;
  }

  res.render('new_appointment.hbs', {
    title: 'New Appointment | IsOn24 ExAPI Client App',
    user: {
      client_id: clid,
      business_id: bid,
    }
  });
};

appController.update_appointment_get = async (req, res) => {
  logger.debug(req.originalUrl);
  let return_obj = { title: 'Update Appointment | IsOn24 ExAPI Client App', };

  try {
    const { clid, bid, appt_id, client_tz } = req.query || {};
    logger.debug(`clid: ${clid}, bid: ${bid}, appt_id: ${appt_id}, client_tz: ${client_tz}`);

    if (isEmptyString(clid) || isEmptyString(bid) || isEmptyString(appt_id)) {
      throw new Error('Empty request');
    }

    return_obj.user = {
      client_id: clid,
      business_id: bid,
    };

    // Get appointment from DB
    let appointment = await exapi_client_model.getAppointment(clid, bid, appt_id);
    if (!appointment) {
      throw new Error(`Appointment ${appt_id} not found`);
    }
    logger.debug(`appointment: ${JSON.stringify(appointment)}`);

    // Fill contact info
    let contact = await exapi_client_model.getContact(clid, bid, { _id: appointment.contact_id });
    if (!contact) {
      throw new Error(`Contact ${appointment.contact_id} not found`);
    }
    logger.debug(`contact: ${JSON.stringify(contact)}`);

    // Fill appointment info
    appointment = {
      ...appointment,
      ContactNumber: formatPhoneNumber(contact.phone),
      ContactName: contact.first_name,
      ContactEmail: contact.email,
      StartTime: getDateFormated(new Date(appointment.start), client_tz),
      Duration: (Date.parse(appointment.end) - Date.parse(appointment.start)) / 1000 / 60,
    };

    return_obj.appointment = appointment;

  } catch (error) {
    return_obj.error = error.toString();
    logger.error(`update_appointment_get error: ${return_obj.error}`);
    // res.status(500).send('Error creating appointment');
  }
  logger.debug(`return_obj: ${JSON.stringify(return_obj)}`);
  res.render('update_appointment.hbs', return_obj);
};

appController.update_appointment_post = async (req, res) => {
  logger.debug(req.originalUrl);

  let return_obj = { title: 'Update Appointment | IsOn24 ExAPI Client App', };

  try {
    const { clid, bid,
      appt_id, contact_phone, contact_name, contact_email,
      date, time, duration, start_ts,
      type, guest_count, content,
      use_get_api, client_tz } = req.body;

    logger.debug(`clid: ${clid}, bid: ${bid}, client_tz: ${client_tz}, appointment_id: ${appt_id}`);
    return_obj.user = {
      client_id: clid,
      business_id: bid,
    };

    logger.debug(`contact_phone: ${contact_phone}, date: ${date}, time: ${time}, duration: ${duration}, start_ts: ${start_ts}`);

    if (isEmptyString(contact_phone) || isEmptyString(date) || isEmptyString(time) || isEmptyString(duration)) {
      throw new Error('Empty request');
    }

    let appointment = {};
    /*
    First, pull the old appointment from DB
    Update the values in the object
    Save updates to appointments to DB
    call ExAPI to create appointment in IsOn24
    */
    const old_appointment = await exapi_client_model.getAppointment(clid, bid, appt_id);
    if (!old_appointment) {
      throw new Error(`Appointment ${appt_id} not found`);
    }
    // Save received appointment update to DB
    appointment = await exapi_client_model.updateAppointment(
      clid, bid,
      {
        _id: appt_id,
        Aid: old_appointment.id_ext,
        ContactNumber: filterPhoneNumberInput(contact_phone),
        ContactName: contact_name,
        ContactEmail: contact_email,

        StartTime: parseInt(start_ts),
        Duration: parseInt(duration),
        EndTime: parseInt(start_ts) + parseInt(duration) * 60 * 1000,

        State: 'CONFIRMED',
        EventType: type,
        GuestCount: 1,
        Content: (content ? content : 'Appointment created by ExAPI Client App'),
        CustomInfo: '{}',
      });

    const appointment_update = {
      ...appointment,
      phone: filterPhoneNumberInput(contact_phone),
      name: contact_name,
      email: contact_email,
    };
    logger.debug(`appointment_update: ${JSON.stringify(appointment_update)}`);
    // Create appointment in IsOn24
    await exapi_client_helper.updateAppointment(clid, bid,
      appointment_update);

    // Refresh appointments page
    // renderCalendarPage(req, res, clid, bid, 10);
    return_obj.result = 'Appointment updated.';
    return_obj.appointment = appointment;

  } catch (error) {
    return_obj.error = error.toString();
    logger.error(`update_appointment_post error: ${return_obj.error}`);
    // res.status(500).send('Error creating appointment');
  }
  logger.debug(`return_obj: ${JSON.stringify(return_obj)}`);
  res.render('update_appointment.hbs', return_obj);
};

/**
 * Renders the calls page
 * @param {*} req 
 * @param {*} res 
 * @returns 
 */
appController.calls = async (req, res) => {
  logger.debug(req.originalUrl);

  const { clid, bid, max_count, use_get_api, client_tz } = req.body;
  logger.debug(`clid: ${clid}, bid: ${bid}, client_tz: ${client_tz}`);

  let calls = [];
  if (use_get_api) {
    /*
    pull calls from ExAPI
    populate the calls page
    */
    calls = await exapi_client_helper.getCallRecords(clid, bid);
    logger.debug(`calls: ${JSON.stringify(calls)}`);

    calls = calls.map((call) => {
      return {
        ...call,
        // ResultType: call.result_type,
        ContactNumberFormated: formatPhoneNumber(call.ContactNumber),
        // ContactName: contacts_map[call.contact_id].first_name,
        BeginTimeFormated: getDateFormated( new Date(call.BeginTime), client_tz),
      };
    });

    res.render('calls.hbs', {
      title: 'Calls | IsOn24 ExAPI Client App',
      calls,
      user: {
        client_id: clid,
        business_id: bid,
      }
    });
    return;
  }
  /*
    pull calls from DB
    populate the calls page
    */
  calls = await exapi_client_model.getCallRecords(clid, bid, 10);
  logger.debug(`calls: ${JSON.stringify(calls)}`);

  if (calls && calls.length > 0) {
    // Fill calls with contacts
    let contacts = await exapi_client_model.getContacts(clid, bid, 20);
    logger.debug(`contacts: ${JSON.stringify(contacts)}`);
    let contacts_map = {};
    contacts.forEach((contact) => {
      contacts_map[contact._id] = contact;
    });
    calls = calls.map((call) => {
      return {
        ...call,
        ResultType: call.result_type,
        ContactNumberFormated: formatPhoneNumber(contacts_map[call.contact_id].phone),
        ContactName: contacts_map[call.contact_id].first_name,
        BeginTimeFormated: getDateFormated( call.BeginTime, client_tz),
      };
    });

    logger.debug(`calls: ${JSON.stringify(calls)}`);
  }

  res.render('calls.hbs', {
    title: 'Calls | IsOn24 ExAPI Client App',
    calls: calls,
    user: {
      client_id: clid,
      business_id: bid,
    }
  });
};

appController.test = async (req, res) => {
  logger.info(req.originalUrl);

  try {
    logObject(req.headers);

    logger.info((typeof req.body));
    logger.info(req.body.toString());
    logObject(req.body);

  } catch (error) {
    logger.error('test error: ', error.toString());
  }

  res.status(200).end();
};

export default appController;