
import simpleOAuthPkg from "simple-oauth2";
const { AuthorizationCode } = simpleOAuthPkg;

import logger from "../logger.js";

import { formatPhoneNumber, isEmptyString, logObject } from "../ison24/appUtils.js";

import exapi_client_helper from "../exapi-client/helper.js";

import exapi_client_model from "../exapi-client/model.js";

/**
* OAuth2 configuration
*/
const ison24_mgmt_app_url = process.env.client_mgmt_app_url ; // + '/dashboard';
const CALLBACK_URL = process.env.client_app_host_url + '/oauth/callback';
const API_AUTH_URL = process.env.API_HOST_URL + '/exapi/oauth/authorize';
const CLIENT_ID = process.env.CLIENT_ID;
const CLIENT_SECRET = process.env.CLIENT_SECRET;
const CLIENT_AUTHORIZATION_TYPE = 'authorization_code';

const config = {
  client: {
    id: CLIENT_ID,
    secret: CLIENT_SECRET
  },
  auth: {
    authorizeHost: process.env.API_HOST_URL,
    authorizePath: process.env.AUTHORIZE_PATH,
    tokenHost: API_AUTH_URL,
    tokenPath: '/exapi/oauth/token'
  }
};

const client = new AuthorizationCode(config);

const oAuth2Controller = {};

/**
* Called from the Initial IsOn24 listing URL
* @param {*} req 
* @param {*} res 
*/
oAuth2Controller.auth = (req, res) => {
  logger.info('/auth ');

  logObject(config, 3);
  logObject(client, 3);

  // Generate the authorization URL
  // const stateStr = getAuthorizationUrlState()
  const stateToken = exapi_client_model.createTimedToken(
    { context: 'ison24-exapi', expires_at: new Date(Date.now() + 1000 * 300) })
    .then((stateToken) => {
      const authorizationUri = client.authorizeURL({
        redirect_uri: CALLBACK_URL,
        // scope: 'read',
        state: stateToken.token,
        // customParam: 'foo', // non-standard oauth params may be passed as well
      });

      logger.info(authorizationUri);
      res.redirect(authorizationUri);

    })
    .catch((reason) => {
      utils.error(reason);
      if (reason.cause) {
        utils.log(`error code: ${reason.cause}`);
        // return { API_RESULT: "NOTFOUND" };
      }
      let err = 'Something went wrong.';
      logger.error(err);
      res.render('connection-error.hbs', { error: err });
      // return res.status(500).end();
    });
};

/**
* Callback service parsing the authorization token and asking for the access token 
* 
* @param {*} req 
* @param {*} res 
* @returns 
*/
oAuth2Controller.callback = async (req, res) => {
  logger.info('/callback ');
  const { code, state } = req.query;
  if (!code) {
    let err = 'You did not authorize the Demo API client app. Please try connecting again.'
    logger.error(err);
    res.render('home.hbs', { error: err });
    return;
  }

  // Store the authorization code
  logger.info('Received authorization code: ' + code);
  logger.info('with authorization state: ' + state);

  try {
    // Verify the state
    if (!isEmptyString(state)) {
      // let isValidState = await isValidAuthorizationUrlState(state);
      let isValidState = await exapi_client_model.isValidTimedToken(
        { context: 'ison24-exapi', token: state });
      if (!isValidState) {
        let err = 'Invalid state received with authorization : ' + state;
        logger.error(err);
        // throw new Error(err);
        res.render('connection-error.hbs', { error: err });
        return;
      }
    }

    // As expected // Get the access and refresh tokens from ExAPI
    let options = {
      code: code,
      client_id: CLIENT_ID,
      client_secret: CLIENT_SECRET,
      grant_type: CLIENT_AUTHORIZATION_TYPE,
      redirect_uri: CALLBACK_URL,
    };

    /**
    * https://developer.ison24.com/docs/getting_started 
    * 
    * Request format 
    POST /exapi/oauth/token
    Host: exapi.ison24.com
    Content-Type: application/x-www-form-urlencoded
    
    client_id=<CLIENT_ID>&client_secret=<CLIENT_SECRET>&grant_type=authorization_code&code=<AUTHORIZATION_CODE>
    
    // For postman
    POST https://exapi.ison24.com/exapi/oauth/token
    
    * Response format 
    The result will look something like this:
    
    {
      "access_token": "eyJhbGciOiJIUzI1NiJ9",
      "refresh_token": "e27ccefbb40de0b78a68a19bb986e279"
    }
        
    */
    const tokenResult = await client.getToken(options);
    logger.info(typeof tokenResult);
    logObject(tokenResult);

    // Not sure why - but we need this hack of converting to String and back to object
    const tokenstr = JSON.stringify(tokenResult);
    logger.info(typeof tokenstr);
    logger.info(tokenstr);

    const tokens = JSON.parse(tokenstr);
    logger.info(typeof tokens['accessToken']);
    logObject(tokens['accessToken']);

    // Store the latest access and refresh tokens 
    await exapi_client_model.saveTokens(tokens);

    // Get the connected business information
    // const result = await testAPI(tokens['accessToken']);
    const result = await exapi_client_helper.getBusiness(tokens['accessToken']);
    logObject(result, 4);

    // make sure there was exactly one business received in the result
    if (!result || !result.businesses || !result.businesses[0]) {
      let err = 'No business found in the ExAPI response';
      logger.error(err);
      res.render('connection-error.hbs', { error: err });
      return;
    }
    // save the connected business information
    await exapi_client_model.saveUser(
      {
        ...result.businesses[0],
        state: 'authorized',
        client_id: CLIENT_ID
      });

    logger.info(`Saved ExApi session ${JSON.stringify(result)} `);
    // Confirm to user that we have connected with IsOn24 ExAPI

    res.render('connected-success.hbs',
      {
        title: 'Connected | IsOn24 ExAPI Client App',
        email: result.businesses[0].emails[0], // req.body.email,
        bus_name: result.businesses[0].name, // regObj.BusinessRecord.Name,
        ison24_number: result.businesses[0].sPhones[0], // formatPhoneNumber(regObj.BusinessRecord.SNumber),
        ison24_app_url: ison24_mgmt_app_url // regObj.ContactRecord.NextUrl
      });
    return;
  } catch (error) {
    logger.error('Access Token Error', error.message);
    let err = "We are having some trouble connecting your Jobber account with your IsOn24 AI voice assistant.";
    res.render('connection-error.hbs', { error: err });
    return;
  }
};

oAuth2Controller.home = async (req, res) => {
  logger.info('/home ');
  res.render('home.hbs', {
    title: 'Home | IsOn24 ExAPI Client App',
  })
};

oAuth2Controller.test = async (req, res) => {
  logger.info(req.originalUrl);

  try {
    logObject(req.headers);

    logger.info((typeof req.body));
    logger.info(req.body.toString());
    logObject(req.body);

  } catch (error) {
    logger.error('test error: ', error.toString());
  }

  res.status(200).end();
};

export default oAuth2Controller;