
import { MongoClient, ServerApiVersion } from 'mongodb';
import { v4 as uuidv4 } from 'uuid';
import mongoose from 'mongoose';
import { Schema } from 'mongoose';

const db_uri = process.env.db_uri ;

import logger from "../logger.js";
import { filterPhoneNumberInput } from '../ison24/appUtils.js';

let exapi_client_model = {
  connection: null,
};

exapi_client_model.init = async function () {
  try {
    logger.info('Connecting to DB...');
    mongoose.set('debug', true);

    // Connect the client to the server	(optional starting in v4.7)
    exapi_client_model.connection = await mongoose.connect(db_uri, {
      bufferCommands: false, // Disable mongoose buffering
      // useNewUrlParser: true,
      // useUnifiedTopology: true,
      // useFindAndModify: false,
      // useCreateIndex: true,
    });

    // Send a ping to confirm a successful connection
    // await exapi_client_model.client.db("admin").command({ ping: 1 });
    // console.log("Pinged your deployment. You successfully connected to MongoDB!");
    logger.info("Connected to MongoDB!");
  }
  catch (error) {
    logger.error(`Error connecting to MongoDB: ${error}`);
  }
  finally {
    // Ensures that the client will close when you finish/error
    // await exapi_client_model.client.close();
  }
}

exapi_client_model.shutdown = async function () {
  try {
    logger.info('Closing DB connection...');
    await exapi_client_model.connection.close();
    logger.info('DB connection closed.');
  }
  catch (error) {
    logger.error(`Error closing DB connection: ${error}`);
  }
};

/**
 * Schema for the ExAPI client
 */
mongoose.model("OAuthApp",
  new Schema({
    _id: { type: String, default: uuidv4 },
    name: { type: String, required: true },
    client_id: { type: String, required: true },
    client_secret: { type: String, required: true },
    redirect_uri: { type: String, required: true },
    created_at: { type: Date, default: Date.now },
    updated_at: { type: Date, default: Date.now },
  }),
  "OAuthApps"
);

const OAuthAuthorizationCodeSchema = new Schema({
  _id: { type: String, default: uuidv4 },

  code: { type: String, required: true },
  client_id: { type: String, required: true },
  redirect_uri: { type: String, required: true },
  user_id: { type: String, required: true },
  scope: { type: String, required: true },
  expires_at: { type: Date, required: true },
  created_at: { type: Date, default: Date.now },
  updated_at: { type: Date, default: Date.now },
});
OAuthAuthorizationCodeSchema.index({ code: 1 }, { unique: true });
OAuthAuthorizationCodeSchema.index({ client_id: 1, user_id: 1 }, { unique: true });

mongoose.model("OAuthAuthorizationCode",
  OAuthAuthorizationCodeSchema,
  "OAuthAuthorizationCodes"
);

const OAuthAccessTokenSchema = new Schema({
  _id: { type: String, default: uuidv4 },
  access_token: { type: String, required: true },
  client_id: { type: String, required: true },
  user_id: { type: String, required: true },
  scope: { type: String, required: true },
  expires_at: { type: Date, required: true },
  created_at: { type: Date, default: Date.now },
  updated_at: { type: Date, default: Date.now },
});
OAuthAccessTokenSchema.index({ access_token: 1 }, { unique: true });
OAuthAccessTokenSchema.index({ client_id: 1, user_id: 1 }, { unique: true });

mongoose.model("OAuthAccessToken",
  OAuthAccessTokenSchema,
  "OAuthAccessTokens"
);

const OAuthRefreshTokenSchema = new Schema({
  _id: { type: String, default: uuidv4 },
  refresh_token: { type: String, required: true },
  client_id: { type: String, required: true },
  user_id: { type: String, required: true },
  scope: { type: String, required: true },
  expires_at: { type: Date, required: true },
  created_at: { type: Date, default: Date.now },
  updated_at: { type: Date, default: Date.now },
});
OAuthRefreshTokenSchema.index({ refresh_token: 1 }, { unique: true });
OAuthRefreshTokenSchema.index({ client_id: 1, user_id: 1 }, { unique: true });

mongoose.model("OAuthRefreshToken",
  OAuthRefreshTokenSchema,
  "OAuthRefreshTokens"
);

const OAuthUsersSchema = new Schema({
  _id: { type: String, default: uuidv4 },
  client_id: { type: String, required: true },
  business_id: { type: String, required: true },
  business_name: { type: String, required: true },
  business_phone: { type: String, required: true },
  business_email: { type: String, required: true },
  business_timezone: { type: String, required: true },
  state: { type: String, required: true, default: "authorized" }, // authorized, active, inactive, deleted

  // username: { type: String, required: false },
  // password: { type: String, required: false },
  // first_name: { type: String },
  // last_name: { type: String },
  // email: { type: String },
  // phone: { type: String },
  // address: { type: String },
  // city: { type: String },
  // state: { type: String },
  // zip: { type: String },
  // country: { type: String },

  created_at: { type: Date, default: Date.now },
  updated_at: { type: Date, default: Date.now },
});
OAuthUsersSchema.index({ client_id: 1, business_id: 1 }, { unique: true });
OAuthUsersSchema.pre('save', function (next) {
  this.updated_at = Date.now();
  return next();
});

mongoose.model("OAuthUser",
  OAuthUsersSchema,
  "OAuthUsers"
);

const OAuthContactsSchema = new Schema({
  _id: { type: String, default: uuidv4 },
  client_id: { type: String, required: true },
  business_id: { type: String, required: true },
  first_name: { type: String },
  last_name: { type: String },
  email: { type: String },
  phone: { type: String },
  created_at: { type: Date, default: Date.now },
  updated_at: { type: Date, default: Date.now },
});
OAuthContactsSchema.index({ client_id: 1, business_id: 1, phone: 1 }, { unique: true });

mongoose.model("Contact",
  OAuthContactsSchema,
  "Contacts"
);

const OAuthAppointmentsSchema = new Schema({
  _id: { type: String, default: uuidv4 },
  client_id: { type: String, required: true },
  business_id: { type: String, required: true },
  contact_id: { type: String, required: true },
  id_ext: { type: String, required: false, unique: false }, // hack to allow for null values
  type: { type: String, required: true },
  start: { type: Date, required: true },
  end: { type: Date, required: true },
  timezone: { type: String, required: true },
  state: { type: String, required: true },
  url: { type: String, required: true },
  notes: { type: String, required: true },
  data_guest_count: { type: Number, required: true, default: 1 },
  data_content: { type: String, required: true },
  data_custom_info: { type: String, required: true },
  created_at: { type: Date, default: Date.now },
  updated_at: { type: Date, default: Date.now },
});
// OAuthAppointmentsSchema.index({ client_id: 1, business_id: 1, contact_id: 1, id_ext: 1 }, { unique: true });
OAuthAppointmentsSchema.index({ client_id: 1, business_id: 1, contact_id: 1, id_ext: 1 }, { unique: false });// hack to allow for null id_ext values

mongoose.model("Appointment",
  OAuthAppointmentsSchema,
  "Appointments"
);

const OAuthMessagesSchema = new Schema({
  _id: { type: String, default: uuidv4 },
  client_id: { type: String, required: true },
  business_id: { type: String, required: true },
  contact_id: { type: String, required: true },
  id_ext: { type: String, required: true, unique: true },
  sender_type: { type: String, required: true },
  sender_id: { type: String, required: true },
  conversation_id: { type: String, required: true },
  message_type: { type: String, required: true, default: "text" },
  message_body: { type: String, required: true },
  created_at: { type: Date, default: Date.now },
  updated_at: { type: Date, default: Date.now },
});
OAuthMessagesSchema.index({ client_id: 1, business_id: 1, contact_id: 1, id_ext: 1 }, { unique: true });

mongoose.model("Message",
  OAuthMessagesSchema,
  "Messages"
);

const OAuthCallRecordsSchema = new Schema({
  _id: { type: String, default: uuidv4 },
  client_id: { type: String, required: true },
  business_id: { type: String, required: true },
  contact_id: { type: String, required: true },
  id_ext: { type: String, required: true, unique: true },
  result_type: { type: String, required: true },
  result_id: { type: String, required: true },
  result_link: { type: String, required: true },
  transcript_link: { type: String, required: true },
  call_length: { type: Number, required: true },
  begin_time: { type: Date, required: true },
  end_time: { type: Date, required: true },
  to_number: { type: String, required: true },
  created_at: { type: Date, default: Date.now },
  updated_at: { type: Date, default: Date.now },
});
OAuthCallRecordsSchema.index({ client_id: 1, business_id: 1, contact_id: 1, id_ext: 1 }, { unique: true });

mongoose.model("CallRecord",
  OAuthCallRecordsSchema,
  "CallRecords"
);

mongoose.model("TimedToken",
  new Schema({
    _id: { type: String, default: uuidv4 },
    token: { type: String, required: true, default: uuidv4 },
    context: { type: String, required: true, default: "default" },
    expires_at: { type: Date, required: true },
    created_at: { type: Date, default: Date.now },
  }),
  "TimedTokens"
);

const OAuthAppsModel = mongoose.model("OAuthApp");
const OAuthAuthorizationCodesModel = mongoose.model("OAuthAuthorizationCode");
const OAuthAccessTokensModel = mongoose.model("OAuthAccessToken");
const OAuthRefreshTokensModel = mongoose.model("OAuthRefreshToken");
const OAuthUsersModel = mongoose.model("OAuthUser");
const ContactsModel = mongoose.model("Contact");
const AppointmentsModel = mongoose.model("Appointment");
const MessagesModel = mongoose.model("Message");
const CallRecordsModel = mongoose.model("CallRecord");
const TimedTokensModel = mongoose.model("TimedToken");

exapi_client_model.createTimedToken = async function (ctx) {
  logger.debug(`createTimedToken: ctx: ${JSON.stringify(ctx)}`);
  try {
    const token = await TimedTokensModel.create({ context: ctx.context, expires_at: ctx.expires_at });
    return token;
  } catch (error) {
    logger.error(`Error creating timed token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.isValidTimedToken = async function (ctx) {
  logger.debug(`isValidTimedToken: ctx: ${JSON.stringify(ctx)}`);
  try {
    const token = await TimedTokensModel.findOne({ token: ctx.token, context: ctx.context });
    if (token) {
      return token.expires_at > Date.now();
    }
    // return token;
  } catch (error) {
    logger.error(`Error getting timed token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.deleteTimedToken = async function (ctx) {
  logger.debug(`deleteTimedToken: ctx: ${JSON.stringify(ctx)}`);
  try {
    const token = await TimedTokensModel.findOneAndDelete({ token: ctx.token, context: ctx.context });
    return token;
  } catch (error) {
    logger.error(`Error deleting timed token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.saveOAuthAccessToken = async function (token, client_id, expires_at, user_id, scope) {
  logger.debug(`saveOAuthAccessToken: token: ${token}, client_id: ${client_id}, expires_at: ${expires_at}, user_id: ${user_id}, scope: ${scope}`);
  try {
    const accessToken = // await OAuthAccessTokensModel.create({ access_token: token, client_id: client_id, expires_at: expires_at, user_id: user_id, scope: scope });
      await OAuthAccessTokensModel.updateOne(
        { client_id: client_id, user_id: user_id },
        { access_token: token, client_id: client_id, expires_at: expires_at, user_id: user_id, scope: scope, updated_at: Date.now() },
        { upsert: true, setDefaultsOnInsert: true });
    return accessToken;
  } catch (error) {
    logger.error(`Error saving OAuth access token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.getOAuthAccessToken = async function (token) {
  logger.debug(`getOAuthAccessToken: token: ${token}`);
  try {
    const accessToken = await OAuthAccessTokensModel.findOne({ access_token: token });
    return accessToken;
  } catch (error) {
    logger.error(`Error getting OAuth access token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.findOAuthAccessToken = async function (client_id, user_id) {
  logger.debug(`findOAuthAccessToken: client_id: ${client_id}, user_id: ${user_id}`);
  try {
    const accessToken = await OAuthAccessTokensModel.findOne({ client_id: client_id, user_id: user_id });
    return accessToken;
  } catch (error) {
    logger.error(`Error finding OAuth access token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.saveOAuthRefreshToken = async function (token, client_id, expires_at, user_id, scope) {
  logger.debug(`saveOAuthRefreshToken: token: ${token}, client_id: ${client_id}, expires_at: ${expires_at}, user_id: ${user_id}, scope: ${scope}`);
  try {
    const refreshToken = // await OAuthRefreshTokensModel.create({ refresh_token: token, client_id: client_id, expires_at: expires_at, user_id: user_id, scope: scope });
      await OAuthRefreshTokensModel.updateOne(
        { client_id: client_id, user_id: user_id },
        { refresh_token: token, client_id: client_id, expires_at: expires_at, user_id: user_id, scope: scope, updated_at: Date.now() },
        { upsert: true, setDefaultsOnInsert: true });

    return refreshToken;
  } catch (error) {
    logger.error(`Error saving OAuth refresh token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.getOAuthRefreshToken = async function (token) {
  logger.debug(`getOAuthRefreshToken: token: ${token}`);
  try {
    const refreshToken = await OAuthRefreshTokensModel.findOne({ refresh_token: token });
    return refreshToken;
  } catch (error) {
    logger.error(`Error getting OAuth refresh token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.findOAuthRefreshToken = async function (client_id, user_id) {
  logger.debug(`findOAuthRefreshToken: client_id: ${client_id}, user_id: ${user_id}`);
  try {
    const refreshToken = await OAuthRefreshTokensModel.findOne({ client_id: client_id, user_id: user_id });
    return refreshToken;
  } catch (error) {
    logger.error(`Error finding OAuth refresh token: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.saveTokens = async function (tokens) {
  /*
   {
  "accessToken": "d55e8fc4879111eebe4c067be461895f",
  "accessTokenExpiresAt": "2023-11-20T11:44:50.568Z",
  "refreshToken": "d5681d26879111eebe4c067be461895f",
  "refreshTokenExpiresAt": "2023-12-20T10:44:50.568Z",
  "scope": "[{\"name\":\"Clients\",\"read_only\":true,\"read_write\":false},{\"name\":\"Messages\",\"read_only\":true,\"read_write\":false},{\"name\":\"Call records\",\"read_only\":true,\"read_write\":false},{\"name\":\"Appointments\",\"read_only\":true,\"read_write\":false}]",
  "client": {
    "id": "CI445a8f6a839311eebe4c067be461895f"
  },
  "user": {
    "id": "49e6ec2b-6ec1-11ee-94ac-067be461895f"
  }
}
  */
  logger.debug(`saveTokens: tokens: ${JSON.stringify(tokens)}`);
  try {
    const accessToken =
      // await OAuthAccessTokensModel.create({ access_token: tokens.accessToken, client_id: tokens.client.id, expires_at: tokens.accessTokenExpiresAt, user_id: tokens.user.id, scope: tokens.scope });
      await OAuthAccessTokensModel.updateOne(
        { client_id: tokens.client.id, user_id: tokens.user.id },
        { access_token: tokens.accessToken, client_id: tokens.client.id, expires_at: tokens.accessTokenExpiresAt, user_id: tokens.user.id, scope: tokens.scope, updated_at: Date.now() },
        { upsert: true, setDefaultsOnInsert: true });

    const refreshToken =
      // await OAuthRefreshTokensModel.create({ refresh_token: tokens.refreshToken, client_id: tokens.client.id, expires_at: tokens.refreshTokenExpiresAt, user_id: tokens.user.id, scope: tokens.scope });
      await OAuthRefreshTokensModel.updateOne(
        { client_id: tokens.client.id, user_id: tokens.user.id },
        { refresh_token: tokens.refreshToken, client_id: tokens.client.id, expires_at: tokens.refreshTokenExpiresAt, user_id: tokens.user.id, scope: tokens.scope, updated_at: Date.now() },
        { upsert: true, setDefaultsOnInsert: true });
    return { accessToken, refreshToken };
  } catch (error) {
    logger.error(`Error saving tokens: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.saveUser = async function (user) {
  /*
  {
  "bus": {
    "id": "49e6ec2b-6ec1-11ee-94ac-067be461895f",
    "name": "Peter Parker @ Red canyon appliance repair",
    "sPhones": [ "7811234567" ],
    "emails": [ "pparker@example.com" ],
    "timezone": "US/Pacific"
  }
}
  */
  logger.debug(`saveUser: user: ${JSON.stringify(user)}`);
  try {
    const userObj = await OAuthUsersModel.updateOne(
      { client_id: user.client_id, business_id: user.id },
      {
        client_id: user.client_id,
        business_id: user.id,
        business_name: user.name,
        business_phone: user.sPhones[0],
        business_email: user.emails[0],
        business_timezone: user.timezone, // 'US/Pacific'
        updated_at: Date.now()
      },
      { upsert: true, setDefaultsOnInsert: true });
    // const userObj = await OAuthUsersModel.create(
    //   { client_id: user.client_id, 
    //     business_id: user.bus.id, 
    //     business_name: user.bus.name, 
    //     business_phone: user.bus.sPhones[0],  
    //     business_email: user.bus.emails[0], 
    //     business_timezone: user.bus.timezone, // 'US/Pacific'
    //   });
    return userObj;
  } catch (error) {
    logger.error(`Error saving user: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.updateUserState = async function (clientId, businessId, state) {
  logger.debug(`updateUserState: client_id: ${clientId}, business_id: ${businessId}, state: ${state}`);
  try {
    const userObj = await OAuthUsersModel.findOneAndUpdate(
      { client_id: clientId, business_id: businessId }, { state: state, updated_at: Date.now() }, { new: true });
    return userObj;
  } catch (error) {
    logger.error(`Error updating user state: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.getOAuthUsers = async function (cur_state) {

  logger.debug(`getOAuthUsers: `);
  try {
    const users = await OAuthUsersModel.find({ state: cur_state },
      null,
      {
        sort: { 'updated_at': -1 }, // limit: 1, 
        // projection: { client_id: 1, business_id: 1, business_name: 1, business_email: 1, business_phone: 1, business_timezone: 1 } 
      });
    logger.debug(`getOAuthUsers: users: ${JSON.stringify(users)}`);

    return users;
  } catch (error) {
    logger.error(`Error getting user: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.saveContact = async function (contact) {
  /*
  { 
    client_id: 'CI445a8f6a839311eebe4c067be461895f',
     business_id: '49e6ec2b-6ec1-11ee-94ac-067be461895f',
    ContactId: 66767, ContactNumber: '7811234567', ContactName: 'Peter Parker', ContactEmail: 'pparker@example.com',
  }
  */
  logger.debug(`saveContact: contact: ${JSON.stringify(contact)}`);
  try {
    let contactObj = await ContactsModel.findOneAndUpdate(
      { client_id: contact.client_id, business_id: contact.business_id, phone: contact.ContactNumber },
      {
        client_id: contact.client_id, business_id: contact.business_id,
        first_name: contact.ContactName, email: contact.ContactEmail,
        phone: filterPhoneNumberInput(contact.ContactNumber), updated_at: Date.now()
      },
      { upsert: true, setDefaultsOnInsert: true, new: true });

    // Hack for adding a contact for the first time
    if (!contactObj) {
      contactObj = await ContactsModel.findOne(
        { client_id: contact.client_id, business_id: contact.business_id, phone: contact.ContactNumber });
    }

    return contactObj;
  } catch (error) {
    logger.error(`Error saving contact: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.getContacts = async function (client_id, business_id, max_count) {
  logger.debug(`getContacts: client_id: ${client_id}, business_id: ${business_id}`);
  try {
    const contacts = await ContactsModel.find({ client_id: client_id, business_id: business_id },
      null,
      {
        sort: { 'updated_at': -1 }, // limit: 1, 
        // projection: { client_id: 1, business_id: 1, business_name: 1, business_email: 1, business_phone: 1, business_timezone: 1 } 
      }).limit(max_count);
    logger.debug(`getContacts: contacts: ${JSON.stringify(contacts)}`);

    return contacts;
  } catch (error) {
    logger.error(`Error getting contacts: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.getContact = async function (client_id, business_id, contact_info) {
  logger.debug(`getContact: client_id: ${client_id}, business_id: ${business_id}, ...contact_info: ${JSON.stringify(contact_info)}`);
  try {
    const contact = await ContactsModel.findOne({ client_id: client_id, business_id: business_id, ...contact_info });
    logger.debug(`getContact: contact: ${JSON.stringify(contact)}`);

    return contact;
  } catch (error) {
    logger.error(`Error getting contact: ${error}`);
    throw new Error(error.message);
  }
};
/*
messages: [{
  id: '46416-66767',
  Content: 'Hello, this is a test message',
  Created: 1685868551000,
  ConvId: 46416,
  SenderType: 'BIZ', // CLIENT
  ContactId: 66767, ContactNumber: '7811234567', ContactName: 'Peter Parker', ContactEmail: 'pparker@example.com',
  UserId: 'bizuser@example.com', UserName: 'Biz user',
  HeaderText: 'Message to Peter Parker',
},]
*/
exapi_client_model.saveMessages = async function (client_id, business_id, messages) {
  logger.debug(`saveMessages: client_id: ${client_id}, business_id: ${business_id}, messages: ${JSON.stringify(messages)}`);
  /*
  - first check on the contact and crate / update as needed
  - then save the messages
  */
  try {

    const contactsObj = await exapi_client_model.saveContact({ client_id: client_id, business_id: business_id, ...messages[0] });
    logger.debug(`saveMessages: contactsObj: ${JSON.stringify(contactsObj)}`);

    const messagesObj = await MessagesModel.insertMany(
      messages.map((message) => {
        return {
          client_id: client_id,
          business_id: business_id,
          contact_id: contactsObj._id,
          id_ext: message.id,
          sender_type: message.SenderType,
          sender_id: (message.SenderType == 'BIZ') ? message.UserId : message.ContactNumber,
          conversation_id: message.ConvId,
          message_type: 'text',
          message_body: message.Content,
          created_at: new Date(message.Created),
          updated_at: new Date(message.Created),
        };
      })
    );
    return messagesObj;
  } catch (error) {
    logger.error(`Error saving messages: ${error}`);
    throw new Error(error.message);
  }
};


exapi_client_model.getMessages = async function (client_id, business_id, message_count) {
  logger.debug(`getMessages: client_id: ${client_id}, business_id: ${business_id}`);
  try {
    const messages = await MessagesModel.find({ client_id: client_id, business_id: business_id },
      null,
      {
        sort: { 'created_at': -1 }, // limit: 1, 
        // projection: { client_id: 1, business_id: 1, business_name: 1, business_email: 1, business_phone: 1, business_timezone: 1 } 
      }).limit(message_count);
    logger.debug(`getMessages: messages: ${JSON.stringify(messages)}`);

    return messages.map(async (message) => {
      const contact = await exapi_client_model.getContact(client_id, business_id, { _id: message.contact_id });
      return {
        // id: message.id_ext,
        contact: contact,
        Content: message.message_body,
        Created: message.created_at,
        CreatedFormatted: message.created_at.toLocaleString(),
        ConvId: message.conversation_id,
        SenderType: message.sender_type,
        UserId: message.sender_id,
        UserName: message.sender_id,
        ContactPhone: contact.phone,
        ContactName: contact.first_name,
        ContactEmail: contact.email,
      }
    });

  } catch (error) {
    logger.error(`Error getting messages: ${error}`);
    throw new Error(error.message);
  }
};

/*
  "call_records": [
    {
      "id": "CAaa492e7ba1baa45f1e75e8b992a9f47e",
      "ToNumber": "7812623692",
      "TranscriptLink": "https://biz.dev01.samvadi.com/#/sn/sdt?id=CAaa492e7ba1baa45f1e75e8b992a9f47e",
      "BeginTime": 1703504281000,
      "EndTime": 1703504281000,
      "BeginTimeFormated": "December 25 2023 03:38:01",
      "EndTimeFormated": "December 25 2023 03:38:01",
      "CallLength": "0s",
      "ResultType": "none",
      "ResultId": 0,
      "Initiator": "CLIENT",
      "HeaderText": "Call received from 4088232343 via IsOn24:",
      "ContactId": "101",
      "ContactNumber": "4088232343",
      "ContactName": "Bruce Wayne",
      "ContactEmail": "vimal@ison24.com",
      "Content": "☎️ Received call for none - 0s. More info @ https://biz.dev01.samvadi.com/#/sn/sdt?id=CAaa492e7ba1baa45f1e75e8b992a9f47e"
    }
  ]
*/

exapi_client_model.getCallRecords = async function (client_id, business_id, call_record_count) {
  logger.debug(`getCallRecords: client_id: ${client_id}, business_id: ${business_id}`);
  try {
    const call_records = await CallRecordsModel.find({ client_id: client_id, business_id: business_id },
      null,
      {
        sort: { 'begin_time': -1 }, // limit: 1, 
        // projection: { client_id: 1, business_id: 1, business_name: 1, business_email: 1, business_phone: 1, business_timezone: 1 } 
      }).limit(call_record_count);
    logger.debug(`getCallRecords: call_records: ${JSON.stringify(call_records)}`);

    return call_records.map( (call_record) => {
      // const contact = await exapi_client_model.getContact(client_id, business_id, { _id: call_record.contact_id });
      return {
        contact_id: call_record.contact_id,
        // contact: contact,
        id: call_record.id_ext,
        ToNumber: call_record.to_number,
        // ContactNumber: contact.phone,
        // ContactName: contact.first_name,
        // ContactEmail: contact.email,
        Content: call_record.result_link,
        BeginTime: call_record.begin_time,
        EndTime: call_record.end_time,
        CallLength: call_record.call_length,
        ResultType: call_record.result_type,
      };
    } );
  } catch (error) {
    logger.error(`Error getting call records: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.saveCallRecords = async function (client_id, business_id, call_records) {
  logger.debug(`saveCallRecords: client_id: ${client_id}, business_id: ${business_id}, call_records: ${JSON.stringify(call_records)}`);
  /*
  - first check on the contact and crate / update as needed
  - then save the call records
  */

  try {
    const contactsObj = await exapi_client_model.saveContact({ client_id: client_id, business_id: business_id, ...call_records[0] });
    logger.debug(`saveCallRecords: contactsObj: ${JSON.stringify(contactsObj)}`);

    const callRecordsObj = await CallRecordsModel.insertMany(
      call_records.map((call_record) => {
        return {
          client_id: client_id,
          business_id: business_id,
          contact_id: contactsObj._id,
          id_ext: call_record.id,
          result_type: call_record.ResultType,
          result_id: call_record.ResultId,
          result_link: call_record.Content,
          transcript_link: call_record.TranscriptLink,
          // strip any trailing 's' from the call length
          call_length: call_record.CallLength.replace(/s$/, ''),
          
          begin_time: new Date(call_record.BeginTime),
          end_time: new Date(call_record.EndTime),
          to_number: call_record.ToNumber,
          created_at: new Date(call_record.BeginTime),
          updated_at: new Date(call_record.EndTime),
        };
      })
    );
    return callRecordsObj;
  } catch (error) {
    logger.error(`Error saving call records: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.createAppointment = async function (client_id, business_id, appointment) {
  logger.debug(`createAppointment: client_id: ${client_id}, business_id: ${business_id}, appointment: ${JSON.stringify(appointment)}`);
  try {
    const contactsObj = await exapi_client_model.saveContact({ client_id: client_id, business_id: business_id, ...appointment });
    logger.debug(`createAppointment: contactsObj: ${JSON.stringify(contactsObj)}`);

    const new_appointment = await AppointmentsModel.create(
      {
        client_id: client_id, business_id: business_id, contact_id: contactsObj._id,
        // id_ext: uuidv4().replace(/-/g, ''),
        type: appointment.EventType, start: appointment.StartTime, end: appointment.EndTime, timezone: 'US/Pacific', // appointment.timezone, 
        state: appointment.State, url: 'url', // appointment.url, 
        notes: appointment.Content,
        data_guest_count: appointment.GuestCount, data_content: appointment.Content,
        data_custom_info: appointment.CustomInfo,
        created_at: Date.now(), updated_at: Date.now()
      });

    return (new_appointment ? new_appointment.toJSON() : null);

  }
  catch (error) {
    logger.error(`Error creating appointment: ${error}`);
    throw new Error(error.message);
  }
};

/**
 * 
 * @param {*} client_id 
 * @param {*} business_id 
 * @param {*} appointment in IsOn24 format
 * @returns 
 */
exapi_client_model.updateAppointment = async function (client_id, business_id, appointment) {
  logger.debug(`updateAppointment: client_id: ${client_id}, business_id: ${business_id}, appointment: ${JSON.stringify(appointment)}`);
  try {
    // First find and create/update the contact
    const contactsObj = await exapi_client_model.saveContact({ client_id: client_id, business_id: business_id, ...appointment });
    logger.debug(`updateAppointment: contactsObj: ${JSON.stringify(contactsObj)}`);

    // Second create an update object in local schema format
    let updateObj = {
      client_id: client_id, business_id: business_id, contact_id: contactsObj._id,
      id_ext: (appointment.Aid? appointment.Aid : null),
      type: appointment.EventType, start: appointment.StartTime, end: appointment.EndTime,
      timezone: 'US/Pacific',
      state: appointment.State, url: 'url', // appointment.url,
      notes: appointment.Content,
      data_guest_count: appointment.GuestCount, data_content: appointment.Content,
      data_custom_info: appointment.CustomInfo,
      updated_at: Date.now()
    };

    let appointmentObj = null;
    if (appointment._id) {
      appointmentObj = await AppointmentsModel.findOneAndUpdate(
        { client_id: client_id, business_id: business_id, _id: appointment._id },
        { ...updateObj, },
        { upsert: true, setDefaultsOnInsert: true, new: true });
    } else if (appointment.ExternalId) {
      appointmentObj = await AppointmentsModel.findOneAndUpdate(
        { client_id: client_id, business_id: business_id, _id: appointment.ExternalId },
        { ...updateObj, },
        { upsert: true, setDefaultsOnInsert: true, new: true });
    } else if (appointment.Aid){
      appointmentObj = await AppointmentsModel.findOneAndUpdate(
        { client_id: client_id, business_id: business_id, id_ext: appointment.Aid },
        { ...updateObj, },
        { upsert: true, setDefaultsOnInsert: true, new: true });
    } else {
      appointmentObj = await AppointmentsModel.create({ ...updateObj, });
    }

    return (appointmentObj ? appointmentObj.toJSON() : null);
  }
  catch (error) {
    logger.error(`Error updating appointment: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.cancelAppointment = async function (client_id, business_id, appointment) {
  logger.debug(`cancelAppointment: client_id: ${client_id}, business_id: ${business_id}, appointment: ${JSON.stringify(appointment)}`);
  try {

    // Second create an update object in local schema format
    let updateObj = {
      client_id: client_id, business_id: business_id, 
      state: appointment.State, 
      updated_at: Date.now()
    };

    let appointmentObj = await AppointmentsModel.findOneAndUpdate(
        { client_id: client_id, business_id: business_id, _id: appointment._id },
        { ...updateObj, },
        { upsert: true, setDefaultsOnInsert: true, new: true });

    return (appointmentObj ? appointmentObj.toJSON() : null);
  }
  catch (error) {
    logger.error(`Error updating appointment: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.getMessagesForContact = async function (client_id, business_id, contact_id, message_count) {
  logger.debug(`getMessagesForContact: client_id: ${client_id}, business_id: ${business_id}, contact_id: ${contact_id}`);
  try {
    const messages = await MessagesModel.find({ client_id: client_id, business_id: business_id, contact_id: contact_id },
      null,
      {
        sort: { 'created_at': -1 }, // limit: 1, 
        // projection: { client_id: 1, business_id: 1, business_name: 1, business_email: 1, business_phone: 1, business_timezone: 1 } 
      }).limit(message_count);
    logger.debug(`getMessagesForContact: messages: ${JSON.stringify(messages)}`);

    return messages;
  } catch (error) {
    logger.error(`Error getting messages for contact: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.getAppointment = async function (client_id, business_id, appointment_id) {
  logger.debug(`getAppointment: client_id: ${client_id}, business_id: ${business_id}, appointment_id: ${appointment_id}`);
  try {
    const appointment = await AppointmentsModel.findOne({ client_id: client_id, business_id: business_id, _id: appointment_id });
    return (appointment ? appointment.toJSON() : null);
  } catch (error) {
    logger.error(`Error getting appointment: ${error}`);
    throw new Error(error.message);
  }
};

exapi_client_model.getAppointments = async function (client_id, business_id, appointment_count) {
  logger.debug(`getAppointments: client_id: ${client_id}, business_id: ${business_id}`);
  try {
    const appointments = await AppointmentsModel.find({ client_id: client_id, business_id: business_id },
      null,
      {
        sort: { 'start': -1 }, // limit: 1, 
        // projection: { client_id: 1, business_id: 1, business_name: 1, business_email: 1, business_phone: 1, business_timezone: 1 } 
      }).limit(appointment_count);
    const str = JSON.stringify(appointments);
    logger.debug(`getAppointments: appointments: ${str}`);

    return JSON.parse(str);
  } catch (error) {
    logger.error(`Error getting appointments: ${error}`);
    throw new Error(error.message);
  }
};


export default exapi_client_model;

