/**
* Helper utils for ExAPI client 
*/
import fetch from 'node-fetch';

const { timingSafeEqual, createHmac } = await import('node:crypto');

import simpleOAuthPkg from "simple-oauth2";
const { AuthorizationCode } = simpleOAuthPkg;

import logger from "../logger.js";

import exapi_client_model from "./model.js";

import { formatPhoneNumber, getTSMillis, isBlankString, logObject } from "../ison24/appUtils.js";

const API_AUTH_URL = process.env.API_HOST_URL + '/exapi/oauth/authorize';
const API_EP_URL = process.env.API_HOST_URL + '/exapi/v2';
const API_VERSION = process.env.API_VERSION;
const CLIENT_ID = process.env.CLIENT_ID;
const CLIENT_SECRET = process.env.CLIENT_SECRET;

export const ACCESS_TOKEN_LIFETIME = 1800; // lesser than 1 hour

export const IsOn24_Appointment_Id_prefix = `aid:`;
export const Client_Appointment_Id_prefix = `caid:`;

const exapi_client_helper = {};

exapi_client_helper.testAPI = async function (access_token) {
  logger.debug(`testAPI: access_token: ${access_token}`);
  // let countRequested = iSessionObj.queryParams.count;

  let params = {
    // ...iSessionObj.queryParams,
    bid: "place-holder-bid",
  }
  const qStr = (new URLSearchParams(params)).toString();

  // let config = IsOn24.getConfig();
  // utils.log(`config: ${JSON.stringify(config)}`);

  const url = API_EP_URL
    + '/test'
    + '?' + qStr;

  // const data = 'API_VERSION=1&API_NAME=LOGOUT&SESSION_ID=' + userInfo.SESSION_ID;
  const requestOptions = {
    method: 'GET',
    // credentials: 'include',
    headers: {
      authorization: `Bearer ${access_token}`,
      // "X-ISON24-EXAPI-VERSION": API_VERSION,
      accept: 'application/json'
      // 'content-type': 'application/x-www-form-urlencoded',
      // cookie: 'JSESSIONID=' + iSessionObj.iSessionId,
      // Origin: config.origin_host,
    }
    // , body: data      
  };
  logger.debug(`url: ${url}`);

  return fetch(url, requestOptions)
    .then(response => response.json())
    .then(data => {
      // let temp = data;
      // utils.log( JSON.stringify(data));
      return data;
    })
    .catch(err => {
      logger.error(err);
      throw new Error(err.message);
    });
};

exapi_client_helper.getBusiness = async function (access_token) {
  logger.debug(`getBusiness: access_token: ${access_token}`);
  // let countRequested = iSessionObj.queryParams.count;

  let params = {
    // ...iSessionObj.queryParams,
    bid: "place-holder-bid",
  }
  const qStr = (new URLSearchParams(params)).toString();

  // let config = IsOn24.getConfig();
  // utils.log(`config: ${JSON.stringify(config)}`);

  const url = API_EP_URL
    + '/businesses'
    + '?' + qStr;

  // const data = 'API_VERSION=1&API_NAME=LOGOUT&SESSION_ID=' + userInfo.SESSION_ID;
  const requestOptions = {
    method: 'GET',
    // credentials: 'include',
    headers: {
      authorization: `Bearer ${access_token}`,
      // "X-ISON24-EXAPI-VERSION": API_VERSION,
      accept: 'application/json'
      // 'content-type': 'application/x-www-form-urlencoded',
      // cookie: 'JSESSIONID=' + iSessionObj.iSessionId,
      // Origin: config.origin_host,
    }
    // , body: data      
  };
  logger.debug(`url: ${url}`);

  return fetch(url, requestOptions)
    .then(response => response.json())
    .then(data => {
      // let temp = data;
      // utils.log( JSON.stringify(data));
      return data;
    })
    .catch(err => {
      logger.error(err);
      throw new Error(err.message);
    });
};

exapi_client_helper.disconnect = async function (client_id, business_id) {
  logger.debug(`disconnect: client_id: ${client_id}, business_id: ${business_id}`);

  try {
    // Get the access token
    const access_token = await getAccessToken(client_id, business_id);
    if (!access_token) {
      let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
      logger.error(err);
      throw new Error(err.message);
    }

    const url = `${API_EP_URL}/businesses/${business_id}`;
    logger.debug(`url: ${url}`);

    const requestOptions = {
      method: 'POST',
      // credentials: 'include',
      headers: {
        authorization: `Bearer ${access_token.access_token}`,
        // "X-ISON24-EXAPI-VERSION": API_VERSION,
        accept: 'application/json',
        'content-type': 'application/json',
      },
      body: JSON.stringify({ action: 'disconnect' }),
    };

    return fetch(url, requestOptions)
      .then(response => response.json())
      .then(data => {
        // let temp = data;
        logger.debug(JSON.stringify(data));
        return data;
      })
      .catch(err => {
        logger.error(err);
        throw new Error(err.message);
      });
  }
  catch (error) {
    logger.error(error);
    throw new Error(error.message);
  }
};

function verify_webhook_signature(signature, payload) {
  const hash = createHmac('sha256', CLIENT_SECRET)
    .update(payload)
    .digest('base64');

  // logger.debug(`hash: ${hash}`);

  // return signature === hash;
  return timingSafeEqual(Buffer.from(signature), Buffer.from(hash));
}

exapi_client_helper.verify_webhook_payload = function (req) {
  // logObject(req.headers);
  /*
  req.headers
  {
  "host": "a22aa7464413.ngrok.app",
  "user-agent": "node-fetch/1.0 (+https://github.com/bitinn/node-fetch)",
  "content-length": "386",
  "accept": "application/json",
  "accept-encoding": "gzip,deflate",
  "content-type": "application/json",
  "x-forwarded-for": "73.231.82.249",
  "x-forwarded-host": "a22aa7464413.ngrok.app",
  "x-forwarded-proto": "https",
  "x-ison24-hmac-sha256": "pf4IqBvtvSmGa4pEaqPwN5BuSrkeewmWGtHFkDx3M2c="
}
  */
  // Extract Header X-Jobber-Hmac-SHA256
  let hmac = req.headers['x-ison24-hmac-sha256'];
  if (!hmac) {
    let err = 'No HMAC found';
    logger.warn(err);
    throw new Error(err.message);
  }
  logger.debug(`HMAC: ${hmac}`);

  // Check for payload
  if (!req.body) {
    let err = 'No body found';
    logger.warn(err);
    throw new Error(err.message);
  }

  const body = req.body;
  // logger.debug((typeof body));
  const body_str = JSON.stringify(body);
  logger.debug(body_str);
  // logger.debug( body.toString() );
  // logObject(body);

  let res = verify_webhook_signature(hmac, body_str);
  if (!res) {
    let err = 'Webhook verification failed';
    logger.warn(err);
    throw new Error(err.message);
  }
  logger.debug(`Webhook verification succeeded`);
  // logger.debug(`${body_str}`);
  return body;
};

/**
 * Return a list of IsOn24 businesses for the given email address
 * @param {*} client_id 
 * @param {*} business_id 
 * @param {*} search_object 
 * @returns 
 */
exapi_client_helper.searchBusinesses = async function (client_id, business_id, search_object) {
  logger.debug(`searchBusinesses: client_id: ${client_id}, business_id: ${business_id}`);

  try {
    // Get the access token
    const access_token = await getAccessToken(client_id, business_id);
    if (!access_token) {
      let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
      logger.error(err);
      throw new Error(err.message);
    }

    let params = {
      // ...iSessionObj.queryParams,
      email: search_object.email,
    }
    const qStr = (new URLSearchParams(params)).toString();

    // let config = IsOn24.getConfig();
    // utils.log(`config: ${JSON.stringify(config)}`);

    const url = API_EP_URL + '/businesses' + '?' + qStr;

    // const data = 'API_VERSION=1&API_NAME=LOGOUT&SESSION_ID=' + userInfo.SESSION_ID;
    const requestOptions = {
      method: 'GET',
      // credentials: 'include',
      headers: {
        authorization: `Bearer ${access_token.access_token}`,
        // "X-ISON24-EXAPI-VERSION": API_VERSION,
        accept: 'application/json'
      }
    };
    logger.debug(`url: ${url}`);

    return fetch(url, requestOptions)
      .then(response => response.json())
      .then(data => {
        // let temp = data;
        // utils.log( JSON.stringify(data));
        return data;
      })
      .catch(err => {
        logger.error(err);
        throw new Error(err.message);
      });
  }
  catch (error) {
    logger.error(error);
    throw new Error(error.message);
  }
};

exapi_client_helper.getMessages = async function (client_id, business_id) {
  logger.debug(`getMessages: client_id: ${client_id}, business_id: ${business_id}`);

  try {
    // Get the access token
    const access_token = await getAccessToken(client_id, business_id);
    if (!access_token) {
      let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
      logger.error(err);
      throw new Error(err.message);
    }

    let params = {
      // sts: toMillis('2021-07-01T00:00:00.000Z'),
      count: 100,
    }
    const qStr = (new URLSearchParams(params)).toString();

    const url = API_EP_URL + '/messages' + '?' + qStr;

    const requestOptions = {
      method: 'GET',
      // credentials: 'include',
      headers: {
        authorization: `Bearer ${access_token.access_token}`,
        // "X-ISON24-EXAPI-VERSION": API_VERSION,
        accept: 'application/json'
      }
    };
    logger.debug(`url: ${url}`);

    return fetch(url, requestOptions)
      .then(response => response.json())
      .then(data => {
        // let temp = data;
        // utils.log( JSON.stringify(data));
        return data;
      })
      .catch(err => {
        logger.error(err);
        throw new Error(err.message);
      });
  }
  catch (error) {
    logger.error(error);
    throw new Error(error.message);
  }
};

exapi_client_helper.getCallRecords = async function (client_id, business_id) {
  logger.debug(`getCallRecords: client_id: ${client_id}, business_id: ${business_id}`);

  try {
    // Get the access token
    const access_token = await getAccessToken(client_id, business_id);
    if (!access_token) {
      let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
      logger.error(err);
      throw new Error(err.message);
    }

    let params = {
      // sts: toMillis('2021-07-01T00:00:00.000Z'),
      count: 100,
    }
    const qStr = (new URLSearchParams(params)).toString();

    const url = API_EP_URL + '/call-records' + '?' + qStr;

    const requestOptions = {
      method: 'GET',
      // credentials: 'include',
      headers: {
        authorization: `Bearer ${access_token.access_token}`,
        // "X-ISON24-EXAPI-VERSION": API_VERSION,
        accept: 'application/json'
      }
    };
    logger.debug(`url: ${url}`);

    return fetch(url, requestOptions)
      .then(response => response.json())
      .then(data => {
        // let temp = data;
        // utils.log( JSON.stringify(data));
        return data;
      })
      .catch(err => {
        logger.error(err);
        throw new Error(err.message);
      });
  }
  catch (error) {
    logger.error(error);
    throw new Error(error.message);
  }
};

exapi_client_helper.getAppointments = async function (client_id, business_id, count) {
  logger.debug(`getappointments: client_id: ${client_id}, business_id: ${business_id}`);

  try {
    // Get the access token
    const access_token = await getAccessToken(client_id, business_id);
    if (!access_token) {
      let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
      logger.error(err);
      throw new Error(err.message);
    }

    let params = {
      // sts: toMillis('2021-07-01T00:00:00.000Z'),
      count: count,
    }
    const qStr = (new URLSearchParams(params)).toString();

    const url = API_EP_URL + '/appointments' + '?' + qStr;

    const requestOptions = {
      method: 'GET',
      // credentials: 'include',
      headers: {
        authorization: `Bearer ${access_token.access_token}`,
        // "X-ISON24-EXAPI-VERSION": API_VERSION,
        accept: 'application/json'
      }
    };
    logger.debug(`url: ${url}`);

    return fetch(url, requestOptions)
      .then(response => response.json())
      .then(data => {
        // let temp = data;
        // utils.log( JSON.stringify(data));
        return data;
      })
      .catch(err => {
        logger.error(err);
        throw new Error(err.message);
      });
  }
  catch (error) {
    logger.error(error);
    throw new Error(error.message);
  }
};

exapi_client_helper.sendMessage = async function (client_id, business_id, message) {
  logger.debug(`getMessages: client_id: ${client_id}, business_id: ${business_id}, message: ${JSON.stringify(message)}`);

  try {
    // Get the access token
    const access_token = await getAccessToken(client_id, business_id);
    if (!access_token) {
      let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
      logger.error(err);
      throw new Error(err.message);
    }

    // const qStr = (new URLSearchParams(params)).toString();
    const url = API_EP_URL + '/messages'; // + '?' + qStr;

    const requestOptions = {
      method: 'POST',
      // credentials: 'include',
      headers: {
        authorization: `Bearer ${access_token.access_token}`,
        // "X-ISON24-EXAPI-VERSION": API_VERSION,
        accept: 'application/json',
        'content-type': 'application/json',
      },
      body: JSON.stringify(message),
    };
    logger.debug(`url: ${url}`);

    return fetch(url, requestOptions)
      .then(response => response.json())
      .then(data => {
        // let temp = data;
        // utils.log( JSON.stringify(data));
        return data;
      })
      .catch(err => {
        logger.error(err);
        throw new Error(err.message);
      });
  }
  catch (error) {
    logger.error(error);
    throw new Error(error.message);
  }
};

exapi_client_helper.updateAppointment = async function (client_id, business_id, appointment) {
  logger.debug(`updateAppointment: client_id: ${client_id}, business_id: ${business_id}, appointment: ${JSON.stringify(appointment)}`);

  return await exapi_client_helper.sendAppointmentUpdate(client_id, business_id, appointment, 'UPDATE');
};

exapi_client_helper.cancelAppointment = async function (client_id, business_id, appointment) {
  logger.debug(`cancelAppointment: client_id: ${client_id}, business_id: ${business_id}, appointment: ${JSON.stringify(appointment)}`);

  return await exapi_client_helper.sendAppointmentUpdate(client_id, business_id, appointment, 'CANCEL');
};

exapi_client_helper.createAppointment = async function (client_id, business_id, appointment) {
  logger.debug(`createAppointment: client_id: ${client_id}, business_id: ${business_id}, appointment: ${JSON.stringify(appointment)}`);

  return await exapi_client_helper.sendAppointmentUpdate(client_id, business_id, appointment, 'CREATE');
};

exapi_client_helper.sendAppointmentUpdate = async function (client_id, business_id, appointment, action) {
  logger.debug(`sendAppointmentUpdate: client_id: ${client_id}, business_id: ${business_id}, appointment: ${JSON.stringify(appointment)}, action: ${action}`);

  try {
    // Get the access token
    const access_token = await getAccessToken(client_id, business_id);
    if (!access_token) {
      let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
      logger.error(err);
      throw new Error(err.message);
    }

    let ison24_appointment = {
      Action: action,
      
      ExternalId: appointment._id,
      
      StartTime: Date.parse(appointment.start),
      EndTime: Date.parse(appointment.end),

      ContactNumber: appointment.phone,
      
      // Optional fields
      Aid: (appointment.id_ext? appointment.id_ext: null),

      Duration: (appointment.duration? appointment.duration: null),

      ContactName: (appointment.name? appointment.name: null),
      ContactEmail: (appointment.email? appointment.email: null),

      GuestCount: (appointment.data_guest_count? appointment.data_guest_count: null),
      Content: (appointment.notes? appointment.notes: null),

      Type: (appointment.type? appointment.type: null),
      State: (appointment.state? appointment.state: null),
      
    };

    // const qStr = (new URLSearchParams(params)).toString();
    const url = API_EP_URL + '/appointments'; // + '?' + qStr;

    const requestOptions = {
      method: 'POST',
      // credentials: 'include',
      headers: {
        authorization: `Bearer ${access_token.access_token}`,
        // "X-ISON24-EXAPI-VERSION": API_VERSION,
        accept: 'application/json',
        'content-type': 'application/json',
      },
      body: JSON.stringify(ison24_appointment),
    };
    logger.debug(`url: ${url}`);

    return fetch(url, requestOptions)
      .then(response => response.json())
      .then(data => {
        // let temp = data;
        // utils.log( JSON.stringify(data));
        return data;
      })
      .catch(err => {
        logger.error(err);
        throw new Error(err.message);
      });
  } catch (error) {
    logger.error(error);
    throw new Error(error.message);
  }
};

/**
 * Get a valid access token for the given client_id and business_id
 * @param {*} client_id 
 * @param {*} business_id 
 * @returns 
 */
export async function getAccessToken(client_id, business_id) {
  logger.debug(`getAccessToken: client_id: ${client_id}, business_id: ${business_id}`);

  try {
    // Get the access token
    let access_token = await exapi_client_model.findOAuthAccessToken(client_id, business_id);
    if (!access_token) {
      let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
      logger.error(err);
      throw new Error(err.message);
    }
    // Check for Access Token expiry and refresh if needed
    if (hasAccessTokenExpired(access_token)) {
      logger.debug(`Access token expired. Refreshing...`);
      const new_access_token = await refreshAccessToken(access_token);
      if (!new_access_token) {
        let err = `No access token found for client_id: ${client_id}, business_id: ${business_id}`;
        logger.error(err);
        throw new Error(err.message);
      }
      access_token = new_access_token;
    }
    return access_token;
  }
  catch (error) {
    logger.error(error);
    throw new Error(error.message);
  }
};

export function hasAccessTokenExpired(access_token) {
  const expiryBufferMillis = 1000 * 60; // 1 minute

  let millis = 0;

  // if( typeof (jSession['expires']) == String  ) {
  //   millis = getTSMillis(jSession['expires'] );
  // } else if (Number.isInteger(jSession['expires'] ) ) {
  //   millis  = jSession['expires'];
  // }
  // logger.debug(typeof jSession['EXPIRES']);
  // logger.debug(`access_token.expires_at: ${typeof access_token.expires_at}`);
  millis = getTSMillis(access_token.expires_at);
  // millis = jSession['EXPIRES'] && jSession['EXPIRES'].getTime();

  // logger.debug(`millis: ${millis} ? ${Date.now()}`);

  return (millis < (Date.now() + expiryBufferMillis))
};

export async function refreshAccessToken(access_token) {

  let config = {
    client: {
      id: CLIENT_ID,
      secret: CLIENT_SECRET
    },
    auth: {
      authorizeHost: process.env.API_HOST_URL,
      authorizePath: process.env.AUTHORIZE_PATH,
      tokenHost: API_AUTH_URL,
      tokenPath: '/exapi/oauth/token'
    }
  };

  try {
    // Pull refresh token as well
    const refresh_token = await exapi_client_model.findOAuthRefreshToken(access_token.client_id, access_token.user_id);
    if (!refresh_token) {
      let err = `No refresh token found for client_id: ${access_token.client_id}, user_id: ${access_token.user_id}`;
      logger.error(err);
      throw new Error(err.message);
    }
    logger.debug(`refresh_token: ${JSON.stringify(refresh_token)}`);

    let client = new AuthorizationCode(config);
    let accessTokenObject = client.createToken({
      access_token: access_token.access_token,
      refresh_token: refresh_token.refresh_token,
    });

    let refreshParams = { // scope: '<scope>',
      client_id: CLIENT_ID,
      client_secret: CLIENT_SECRET,
    };

    let newAccessToken = await accessTokenObject.refresh(refreshParams);
    logger.debug(`newAccessToken: ${JSON.stringify(newAccessToken)}`);

    newAccessToken = getNormalizedTokenObject(newAccessToken);
    // {
    //   "accessToken": "b92321f093d811eebe4c067be461895f",
    //   "accessTokenExpiresAt": "2023-12-06T02:42:31.404Z",
    //   "refreshToken": "b92a1b3793d811eebe4c067be461895f",
    //   "refreshTokenExpiresAt": "2024-01-05T01:42:31.404Z",
    //   "scope": "[\"business_read\",\"business_read_write\",\"client_read\",\"message_read\",\"call_record_read\",\"appointment_read\"]",
    //   "client": {
    //     "id": "CIc71a0f468e8811eebe4c067be461895f"
    //   },
    //   "user": {
    //     "id": "49e6ec2b-6ec1-11ee-94ac-067be461895f"
    //   }
    // }
    logger.debug(`normalized newAccessToken: ${JSON.stringify(newAccessToken)}`);
    // Store the latest access and refresh tokens 
    await exapi_client_model.saveTokens(newAccessToken);

    return {
      access_token: newAccessToken.accessToken,
      expires_at: newAccessToken.accessTokenExpiresAt,
      scope: newAccessToken.scope,
      client_id: newAccessToken.client.id,
      user_id: newAccessToken.user.id,
    };
  } catch (error) {
    return { error: `Error refreshing access token: ${error.message}` }
  }
};

export function getNormalizedTokenObject(oauthTokenObj) {
  // Not sure why - but we need this hack of converting to String and back to object
  const tokenstr = JSON.stringify(oauthTokenObj);
  logger.debug(typeof tokenstr);
  logger.debug(tokenstr);

  return JSON.parse(tokenstr);
};

export default exapi_client_helper;
