
import express from "express";
import hbs from 'hbs';

import path from 'path';
import { fileURLToPath } from 'url';

import morganMW from "./morgan-mw.js";

import oAuth2Routes from './routes/oAuth2Routes.js';
import appRoutes from './routes/appRoutes.js';
import webhookRoutes from './routes/webhookRoutes.js';

import {logInitInfo} from "./ison24/appUtils.js";
import Model from "./exapi-client/model.js";
import logger from "./logger.js";

try {
  const __filename = fileURLToPath(import.meta.url);
  const __dirname = path.dirname(__filename);

  const app = express();
  const PORT = process.env.client_app_local_port || 3000;

  app.set('view engine', hbs);
  app.set('views', path.join(__dirname, '/views'));
  hbs.registerPartials(path.join(__dirname, '/views/partials'));

  // Initialize Model
  await Model.init();

  // Print some initialization info
  logInitInfo();

  app.use(express.json({ type: 'application/json' }));
  app.use(express.urlencoded({ extended: true, type: 'application/x-www-form-urlencoded' }));
  app.use(express.static(path.join(__dirname, '/public')))
  app.use(morganMW);

  /** Always add router after adding middlewares. */

  app.use('/', appRoutes);
  app.use('/oauth/', oAuth2Routes);
  app.use('/webhook/', webhookRoutes);

  logger.info(`Starting app on port ${PORT}`);
  const server = app.listen(PORT, () => {
    console.log(`app is running on PORT ${PORT}`)
  })

  process.on('SIGTERM', () => {
    logger.info('SIGTERM signal received.');
    logger.info('Closing http server.');
    server.close(() => {
      logger.info('Http server closed.');
      Model.shutdown();
      process.exit(0);
    });
  });

} catch (error) {
  logger.error(`Main error: ${error}`);
  process.exit(1);
}
