
import logger from "../logger.js";

import exapi_client_helper from '../exapi-client/helper.js';

import { filterPhoneNumberInput, logObject } from "../ison24/appUtils.js";
import exapi_client_model from "../exapi-client/model.js";

const webhookController = {};

const CLIENT_ID = process.env.CLIENT_ID;

/**
 * Main webhook handler
 * @param {*} req 
 * @param {*} res 
 */
webhookController.webhook = async (req, res) => {
  logger.debug(req.originalUrl);

  try {
    // logger.debug(JSON.stringify(req.headers));
    // logger.debug(JSON.stringify(req.body));

    const body = exapi_client_helper.verify_webhook_payload(req);

    // logger.debug('Webhook verified');
    // logger.debug(JSON.stringify(body));
    if(! body?.data?.webHookEvent?.topic) {
      logger.warn(`No webhook event found`);
      res.status(200).end();
      return;
    }
    /*
{"data": {
    "webHookEvent": {
      "topic": "app_connect",
      "appId": "CIc71a0f468e8811eebe4c067be461895f",
      "businessId": "49e6ec2b-6ec1-11ee-94ac-067be461895f",
      "occuredAt": "2023-12-03T00:59:14.317Z",
      "payload": {}
    }
  }
}
    */
    switch (body.data.webHookEvent.topic) {
      case 'app_connect': {
        logger.debug(`app_connect for account ${body.data.webHookEvent.appId} and item ${body.data.webHookEvent.businessId}`);

        await exapi_client_model.updateUserState(
          body.data.webHookEvent.appId,
          body.data.webHookEvent.businessId,
          'active');
        break;
      }
      case 'app_disconnect': {
        logger.debug(`app_disconnect for account ${body.data.webHookEvent.appId} and item ${body.data.webHookEvent.businessId}`);

        await exapi_client_model.updateUserState(
          body.data.webHookEvent.appId,
          body.data.webHookEvent.businessId,
          'inactive');
        break;
      }
      case 'message_received' : {
        /*
        payload: {
  "id": "49e6ec2b-6ec1-11ee-94ac-067be461895f",
  "contact_id": "4081234567",
  "messages": [
    {
      "id": "7329-66394",
      "Content": "💬 [Peter Parker-ison24 sent] ack 1 for api test",
      "Created": 1701933389000,
      "CreatedFormated": "December 06 2023 23:16:29",
      "ConvId": 7329,
      "SenderType": "BIZ",
      "ContactId": "102",
      "ContactNumber": "4081234567",
      "ContactName": "Peter Parker",
      "ContactEmail": "",
      "UserId": "pparker@example.com",
      "UserName": "Peter Parker-ison24",
      "HeaderText": "Message sent by Peter Parker-ison24 via IsOn24:"
    }
  ]
}
        */
        logger.debug(`message_received for account ${body.data.webHookEvent.appId} and item ${body.data.webHookEvent.businessId}`);
        // Save received messages to DB
        await exapi_client_model.saveMessages(
          body.data.webHookEvent.appId, body.data.webHookEvent.businessId, 
          body.data.webHookEvent.payload.messages);
        break;
      }
      case 'call_received' : {
        /*
{
  "id": "49e6ec2b-6ec1-11ee-94ac-067be461895f",
  "contact_id": "4088232343",
  "call_records": [
    {
      "id": "CAaa492e7ba1baa45f1e75e8b992a9f47e",
      "ToNumber": "7812623692",
      "TranscriptLink": "https://biz.dev01.samvadi.com/#/sn/sdt?id=CAaa492e7ba1baa45f1e75e8b992a9f47e",
      "BeginTime": 1703504281000,
      "EndTime": 1703504281000,
      "BeginTimeFormated": "December 25 2023 03:38:01",
      "EndTimeFormated": "December 25 2023 03:38:01",
      "CallLength": "0s",
      "ResultType": "none",
      "ResultId": 0,
      "Initiator": "CLIENT",
      "HeaderText": "Call received from 4088232343 via IsOn24:",
      "ContactId": "101",
      "ContactNumber": "4088232343",
      "ContactName": "Bruce Wayne",
      "ContactEmail": "vimal@ison24.com",
      "Content": "☎️ Received call for none - 0s. More info @ https://biz.dev01.samvadi.com/#/sn/sdt?id=CAaa492e7ba1baa45f1e75e8b992a9f47e"
    }
  ]
}        
        */
        logger.debug(`call_received for account ${body.data.webHookEvent.appId} and item ${body.data.webHookEvent.businessId}`);
        logger.debug(JSON.stringify(body.data.webHookEvent.payload));
        // Save received call to DB
        await exapi_client_model.saveCallRecords(
          body.data.webHookEvent.appId, body.data.webHookEvent.businessId,
          body.data.webHookEvent.payload.call_records);

        break;
      }
      case 'appointment_create' : 
      case 'appointment_update' : 
      case 'appointment_destroy' : {
        /*
                  payload: {
  "id": "49e6ec2b-6ec1-11ee-94ac-067be461895f",
  "contact_id": "4081234567",
         "appointment": {
  "id": String # The unique DE-DUP identifier for a specific update on this appointment
  "Aid": String # The unique identifier for this appointment
  "Created": Integer # Millisecs since Epoch 
  "Updated": Integer # Millisecs since Epoch 
  "EventType": String # "Appointment", "Reservation", "Booking", "Meeting", etc.
  "State": String # "NEW", "CONFIRMED", "CANCELLED", "DECLINED"
  "StartTime": Integer # Millisecs since Epoch 
  "StartTimeFormated": String # Formatted date time in the Business Timezone
  "EndTime": Integer # Millisecs since Epoch 
  "EndTimeFormated": String # Formatted date time in the Business Timezone
  "Duration": Integer # Minutes
  "GuestCount": Integer # 
  "Content": String # Stringified JSON for optional additional attributes
  "ContactId": String # The unique identifier
  "ContactNumber": String # Phone number 
  "ContactName": String # client name 
  "ContactEmail": String # email address
  "HeaderText": String # Optional header text for the appointment
}
      }
      */
      logger.debug(`appointment_update for account ${body.data.webHookEvent.appId} and item ${body.data.webHookEvent.businessId}`);
      // Save received appointment update to DB
      await exapi_client_model.updateAppointment(
        body.data.webHookEvent.appId, body.data.webHookEvent.businessId, 
        body.data.webHookEvent.payload.appointment);

    break;
  }
        
      default:
        logger.warn(`Unknown webhook topic ${body.data.webHookEvent.topic}`);
    }
    // Try to push this webhook event to DB 
    // Spurious events will be rejected by the DB
    // const event = body.data.webHookEvent;

    // await inbound_webhook_event_try_enqueue(Jobber_integration_name,
    //   `${event.topic}-${event.accountId}-${event.itemId}-${event.occuredAt}`,
    //   JSON.stringify(event));
  }
  catch (error) {
    // logger.error(`webhook_client error: ${JSON.stringify( error) }`);
    logger.warn(`webhook: ${JSON.stringify(error) }`);
  }

  res.status(200).end();
};

webhookController.home = async (req, res) => {
  logger.info('/home ');

  res.render('home.hbs')
};

webhookController.webhook_test = async (req, res) => {
  logger.info('/webhook_test ');

  try {
    logObject(req.headers);

    // Log the affected client info
    // const body = { ...req.body };
    // const body = JSON.parse( req.body) ;
    logger.info((typeof req.body));
    logger.info(req.body.toString());
    logObject(req.body);

  } catch (error) {
    logger.error('webhook_test error: ', error.toString());
  }

  res.status(200).end();
};

export default webhookController;